""" XMPP bot. """

__version__ = 2

import _thread
import logging
import ssl
import threading

import bot
import obj

from obj.utils import stripped

try:
    import sleekxmpp
except ImportError:
    pass

def init():
    xmpp = XMPP()
    xmpp.start()
    return xmpp

class Cfg(obj.Cfg):

    def __init__(self):
        super().__init__()
        self.ipv6 = False
        self.nick = ""
        self.noresolver = True
        self.password = ""
        self.channel = ""
        self.server = ""

class XEvent(obj.event.Event):

    def __init__(self):
        super().__init__()
        self.cc = ""
        self.channel = ""
        self.element = ""
        self.jid = ""
        self.nick = ""
        self.server = ""
        self.mtype = ""
        self.txt = ""

class XMPP(bot.Bot):

    def __init__(self):
        super().__init__()
        self._connected = threading.Event()
        self._threaded = False
        self.cc = ""
        self.cfg = Cfg()
        self.channels = []
        self.client = None
        self.orig = repr(self)
        self.jid = None
        self.state = obj.Object()
        self.users = []

    def _connect(self, user, pw):
        logging.warn("connect %s" % user)
        self._makeclient(user, pw)
        if self.cfg.noresolver:
            self.client.configure_dns(None)
        self.client.connect(use_tls=True)
        self._connected.set()
        
    def _makeclient(self, jid, password):
        self.client = sleekxmpp.clientxmpp.ClientXMPP(jid,
                                                      password,
                                                      plugin_config={},
                                                      plugin_whitelist=[],
                                                      escape_quotes=False,
                                                      sasl_mech=None)
        self.client._error = obj.Object()
        self.client.register_plugin(u'xep_0045')
        self.client.add_event_handler('errored', self.handler)
        self.client.add_event_handler('failed_auth', self.handler)
        self.client.add_event_handler("message", self.messaged)
        self.client.add_event_handler("iq", self.handler)
        self.client.add_event_handler('presence', self.presenced)
        self.client.add_event_handler('presence_dnd', self.presenced)
        self.client.add_event_handler('presence_xa', self.presenced)
        self.client.add_event_handler('presence_available', self.presenced)
        self.client.add_event_handler('presence_chat', self.presenced)
        self.client.add_event_handler('presence_away', self.presenced)
        self.client.add_event_handler('presence_unavailable', self.presenced)
        self.client.add_event_handler('presence_subscribe', self.presenced)
        self.client.add_event_handler('presence_subscribed', self.presenced)
        self.client.add_event_handler('presence_unsubscribe', self.presenced)
        self.client.add_event_handler('presence_unsubscribed', self.presenced)
        self.client.add_event_handler("session_bind", self.session_bind)
        self.client.add_event_handler("session_start", self.session_start)
        self.client.add_event_handler("ssl_invalid_cert", self.handler)
        #self.client.add_event_handler("ssl_cert", self.handler)
        self.client.exception = self.handler
        self.client.reconnect_max_attempts = 3
        self.client.ssl_version = ssl.PROTOCOL_SSLv23
        self.client.use_ipv6 = self.cfg.ipv6

    def loop(self):
        self._connected.wait()
        self.client.process(block=True)

    def announce(self, txt):
        for channel in self.channels:
            self.say(self.orig, channel, txt, "chat")
        for origin in self.users:
            self.say(self.orig, origin, txt, "chat")

    def connect(self, user, password):
        assert user, password
        try:
            self._connect(user, password)
            return True
        except sleekxmpp.exceptions.XMPPError as ex:
            logging.error(ex)

    def handler(self, data):
        logging.error(data)

    def do_join(self):
        c = self.cfg
        logging.warning("join %s as %s" % (c.channel, c.nick))
        if c.channel not in self.channels:
            self.channels.append(c.channel)
        self._connected.wait()
        self.client.plugin['xep_0045'].joinMUC(c.channel,
                                               c.nick,
                                               wait=True)

    def fileno(self):
        return self.client.filesocket.fileno()

    def messaged(self, data):
        if '<delay xmlns="urn:xmpp:delay"' in str(data):
            txt = "ignore %s" % data["From"]
            logging.warning(txt)
            return
        origin = str(data["from"])
        if data["type"] == "groupchat":
            if not users.allowed(origin, "USER"):
                return
        txt = data["body"]
        txt = self.get_aliased(txt)
        m = XEvent(txt)
        m.jid = origin
        m.origin = origin
        m.orig = repr(self)
        m.mtype = data["type"]
        m.txt = txt
        if m["mtype"] == "error":
            txt = "error %s" % m.error
            logging.error(txt)
            return
        m.nick = m.origin.split("/")[-1]
        m.user = self.cfg.server
        m.jid = self.cfg.server.split("@")[-1]
        m.channel = m.origin
        m.channel = stripped(m.origin)
        if self.cfg.server == stripped(m.server):
            txt = "ignore %s %s" % (m.mtype, m["from"])
            logging.warning(txt)
            return
        if m.origin not in self.users:
            self.users.append(m.origin)
        bot.run.kernel.put(m)

    def pinged(self, event):
        logging.info(event)

    def presenced(self, data):
        o = XEvent()
        o.mtype = data["type"]
        o.orig = repr(self)
        o.cc = ""
        o.origin = str(data["from"])
        o.jid = o.origin
        o.nick = o.origin.split("/")[-1]
        o.server = self.cfg.server
        o.channel = stripped(o.origin)
        if "txt" not in o:
            o.txt = ""
        o.element = "presence"
        if self.cfg.server in str(data["from"]):
            txt = "ignore %s" % o.origin
            logging.warning(txt)
            return
        if o.mtype == 'subscribe':
            pres = XEvent({'to': o.origin, 'type': 'subscribed'})
            self.client.send_presence(pres)
            pres = XEvent({'to': o.origin, 'type': 'subscribe'})
            self.client.send_presence(pres)
            self.users.append(o.origin)
        elif o.mtype == "unsubscribe":
            if o.origin in self.channels:
                self.channels.remove(o.origin)
            if o.origin in self.users:
                self.users.remove(o.origin)
            return
        elif o.mtype == "available":
            if o.origin not in self.users:
                self.users.append(o.origin)
        elif o.mtype == "unavailable":
            if o.origin in self.users:
                self.users.remove(o.origin)
        logging.warning("presence (%s) %s" % (o.mtype, o.origin))

    def say(self, orig, channel, txt, mtype="chat"):
        try:
            sleekxmpp.jid.JID(channel)
        except sleekxmpp.jid.InvalidJID:
            txt = "wrong jid %s" % channel
            logging.warning(txt)
            return
        if self.cfg.server in channel:
            logging.warning("not writing to self.")
            return
        if channel in self.channels:
            mtype = "groupchat"
        if mtype == "groupchat":
            channel = stripped(channel)
        self.client.send_message(channel, str(txt), mtype)

    def session_bind(self, data):
        self.jid = str(data)
        logging.warning("jid is %s" % self.jid)

    def session_start(self, data):
        logging.info("session start")
        try:
            self.client.send_presence()
            self.client.get_roster()
        except sleekxmpp.exceptions.IqTimeout:
            logging.error("iq timeout %s" % self.cfg.server)
            self.reconnect()
        except sleekxmpp.exceptions.XMPPError as ex:
            logging.error(ex)

    def stop(self):
        if "client" in self and self.client:
            self.client.disconnect()
        super().stop()

    def start(self):
        prev = Cfg()
        prev.update(self.cfg)
        self.cfg.update(obj.cfg)
        if prev != self.cfg:
            self.cfg.save()
        else:
            last = kernel.last("xmppbot.Cfg")
            if last:
                self.cfg.update(last)
        if not self.cfg.server:
            logging.error("use -s to provide a JID")
            return False
        if not self.cfg.password:
            logging.error("use -p to provide a password")
            return False
        ok = self.connect(self.cfg.server, self.cfg.password)
        if ok:
            if self.cfg.channel:
                self.join()
        super().start()
        return self
